-- =====================================================================
-- Copyright © 2010-2012 by Cryptographic Engineering Research Group (CERG),
-- ECE Department, George Mason University
-- Fairfax, VA, U.S.A.
-- =====================================================================

library ieee;
use ieee.std_logic_1164.all;	   
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
use work.sha3_pkg.all;
use work.sha3_skein_package.all;

entity skein_fsm2 is  
	generic ( 
		ppl : integer := 5;
		UF : integer := 4 
	);
	port (
		-- global
		clk : in std_logic;
		rst : in std_logic;
		
		-- datapath
		load_hash, lmsg, clr_msg : out std_logic_vector(ppl-1 downto 0);
		snb, shalf : out std_logic;
		er, etweak,load_tweak  : out std_logic; 	
		sfinal, slast, sf : out std_logic_vector(ppl-1 downto 0);
		stweak, smsg : out std_logic_vector(log2(ppl)-1 downto 0);
		-- control				   
			--fsm1 hand shake signals
		block_ready_clr : out std_logic_vector(ppl-1 downto 0);		
		msg_end_clr 	: out std_logic_vector(ppl-1 downto 0);
		
		block_ready		: in std_logic_vector(ppl-1 downto 0);
		msg_end 		: in std_logic_vector(ppl-1 downto 0);	
				
			--fsm3 handshake signals
		output_busy_set  : out std_logic_vector(ppl-1 downto 0);
		output_busy		 : in  std_logic_vector(ppl-1 downto 0)
	);							   
end skein_fsm2;

architecture beh of skein_fsm2 is  

	constant total_roundnr	: integer := 72;		
	constant roundnr 		: integer := total_roundnr/UF + 1;
	constant log2roundnr 	: integer := log2( roundnr ) + 1;
	constant log2roundnrzeros	: std_logic_vector(log2roundnr-1 downto 0) := (others => '0');
	
	signal round : std_logic_vector(log2roundnr-1 downto 0);
	signal ziroundnr, ziround0, ei, li : std_logic;
	signal snb_sig : std_logic;
	signal output_data_s, block_ready_clr_sig : std_logic_vector(ppl-1 downto 0);	
	signal valid_set, valid_clr, valid, valid_reg : std_logic_vector(ppl-1 downto 0);
	signal sf_clr, sf_set, sf_sig : std_logic_vector(ppl-1 downto 0);
	signal ppl_stage_cnt : std_logic_vector(log2(ppl)-1 downto 0);
	signal block_ready_true : std_logic_vector(ppl-1 downto 0);
	
	signal fb_clr, fb_set, fb : std_logic_vector(ppl-1 downto 0);
	signal done_clr, done_set, done_sig : std_logic_vector(ppl-1 downto 0);
	
	type smsg_type is array ( 0 to ppl-2 ) of std_logic_vector(log2(ppl)-1 downto 0);
	signal smsg_reg : smsg_type;
begin
	-- internals
	pplstage_cnt: process ( clk )
	begin
		if rising_edge( clk ) then		
			if rst = '1' or ei = '1' or li = '1' then
				ppl_stage_cnt <= (others => '0');
			else
				ppl_stage_cnt <= ppl_stage_cnt + 1;
			end if;
		end if;
	end process;   
	
	gsmsg : process(clk)
	begin
		if rising_edge( clk ) then
			smsg_reg(0) <= ppl_stage_cnt;	 
			if ( ppl > 2 ) then
				for i in 1 to ppl-2 loop
					smsg_reg(i) <= smsg_reg(i-1);
				end loop;	   
			end if;
		end if;
	end process;	
	smsg <= smsg_reg(ppl-2);
	stweak <= ppl_stage_cnt;
	
	proc_counter_gen : countern 
		generic map ( N => log2roundnr ) 
		port map ( clk => clk, rst => rst, load => li, en => ei, input => log2roundnrzeros, output => round);
	ziroundnr <= '1' when round = roundnr-1 else '0';	
	ziround0 <= '1' when round = 0 else '0';
	
	
	li <= '1' when (ziroundnr = '1' and ppl_stage_cnt = ppl-1) else '0';
	ei <= '1' when (ppl_stage_cnt = ppl-1) else '0';	
	
	gg1: for i in 0 to ppl-1 generate	 				
		sf_gen : sr_reg 
			generic map ( init => '1' )
			port map ( rst => rst, clk => clk, set => sf_set(i), clr => sf_clr(i), output => sf_sig(i));
		
		done_gen : sr_reg 
			generic map ( init => '0' )
			port map ( rst => rst, clk => clk, set => done_set(i), clr => done_clr(i), output => done_sig(i));
		
		final_block_gen : sr_reg
			generic map ( init => '0' )
			port map ( rst => rst, clk => clk, set => fb_set(i), clr => fb_clr(i), output => fb(i));		
			
		valid_gen : sr_reg 
			generic map ( init => '0' )
			port map ( rst => rst, clk => clk, set => valid_set(i), clr => valid_clr(i), output => valid(i));		
			
		slast(i) <= '1' when (ziroundnr = '1' and msg_end(i) = '1' and fb(i) = '0') else '0';		
		
		fb_set(i) <= '1' when (snb_sig = '1' and smsg_reg(ppl-2) = conv_std_logic_vector((i),log2(ppl)) and msg_end(i) = '1' and fb(i) = '0') else '0';
		fb_clr(i) <= '1' when (snb_sig = '1' and smsg_reg(ppl-2) = conv_std_logic_vector((i),log2(ppl)) and fb(i) = '1') else '0';		
		done_clr(i) <= '1' when (snb_sig = '1' and smsg_reg(ppl-2) = conv_std_logic_vector((i),log2(ppl)) and done_sig(i) = '1') else '0';
	end generate;						
	sf_set <= fb_clr;
	done_set <= fb_clr;
	msg_end_clr <= fb_set;
	
	sf <= sf_sig;
	sfinal <= fb;
	clr_msg <= fb_clr;
	load_tweak <= ziroundnr;
	snb_sig <= '1' when (ziroundnr = '1' and ppl_stage_cnt = ppl-1) or 
						(ziround0 = '1' and ppl_stage_cnt /= ppl-1 ) else '0';
	snb <= snb_sig;
	er <= '1';
	etweak <= '1';
	shalf <= round(0);
	
	
	-- =========================================
	-- Valid and SF	   		
	output_busy_set <= output_data_s;
	load_hash		<= output_data_s;	   	
	
	ctrlsig: for i in 0 to ppl-1 generate		
		output_data_s(i) <= '1' when (ziroundnr = '1' and ppl_stage_cnt = conv_std_logic_vector((i),log2(ppl)) and done_sig(i) = '1' and output_busy(i) = '0' and valid_reg(i) = '1' ) else '0';		
		valid_set(i) <= '1' when (snb_sig = '1' and smsg_reg(ppl-2) = conv_std_logic_vector((i),log2(ppl)) and block_ready_true(i) = '1') else '0';
		valid_clr(i) <= '1' when (snb_sig = '1' and smsg_reg(ppl-2) = conv_std_logic_vector((i),log2(ppl)) and block_ready_true(i) = '0') else '0';
		block_ready_clr_sig(i) <= '1' when (snb_sig = '1' and smsg_reg(ppl-2) = conv_std_logic_vector((i),log2(ppl)) and block_ready_true(i) = '1' and fb(i) = '0') else '0';
	end generate;																			 
	lmsg <= block_ready_clr_sig;
	block_ready_clr <= block_ready_clr_sig;
	sf_clr <= blocK_ready_clr_sig;
	
	vreg : process ( clk ) 
	begin
		if rising_edge( clk ) then	 
			if rst = '1' then
				valid_reg <= (others => '0');				
			elsif li = '1' then  				
				valid_reg <= valid;
			end if;
		end if;
	end process;	
	
	BlockReadyTrueReg : process ( clk ) 
	begin
		if rising_edge( clk ) then	 
			if rst = '1' then
				block_ready_true <= (others => '0');				
			elsif (ziroundnr = '1' and ppl_stage_cnt = ppl-2) then  				
				block_ready_true <= block_ready;
			end if;
		end if;
	end process;
end beh;
		
		
		